import factory
from django.contrib.auth import get_user_model

from tests.models import ExtraData, Profile

User = get_user_model()


class UserFactory(factory.django.DjangoModelFactory):
    uuid = factory.Faker("uuid4", cast_to=None)
    first_name = factory.Faker("first_name")
    last_name = factory.Faker("last_name")
    email = factory.Faker("email")

    class Meta:
        model = User


class ProfileFactory(factory.django.DjangoModelFactory):
    user = factory.SubFactory(UserFactory)
    memo = "Memo"

    class Meta:
        model = Profile


class ExtraDataFactory(factory.django.DjangoModelFactory):
    profile = factory.SubFactory(ProfileFactory)
    data = "Extra"

    class Meta:
        model = ExtraData



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots["test_get_profile_information_from_gdpr_api 1"] = {
    "children": [
        {"key": "MEMO", "value": "Memo"},
        {"key": "USER", "value": "First"},
        {"children": [], "key": "EXTRA_DATA"},
    ],
    "key": "PROFILE",
}



import pytest
from rest_framework.test import APIClient

from tests.factories import ProfileFactory, UserFactory


@pytest.fixture(autouse=True)
def autouse_db(db):
    pass


@pytest.fixture
def api_client():
    return APIClient()


@pytest.fixture
def user():
    return UserFactory()


@pytest.fixture
def profile():
    return ProfileFactory()



import datetime

import pytest
from django.contrib.auth import get_user_model
from django.urls import reverse
from helusers.settings import api_token_auth_settings
from jose import jwt

from .keys import rsa_key
from .models import Profile

User = get_user_model()

TRUE_VALUES = ["true", "True", "TRUE", "1", 1, True]
FALSE_VALUES = ["false", "False", "FALSE", "0", 0, False]


def get_api_token_for_user_with_scopes(user, scopes: list, requests_mock):
    """Build a proper auth token with desired scopes."""
    audience = api_token_auth_settings.AUDIENCE
    issuer = api_token_auth_settings.ISSUER
    auth_field = api_token_auth_settings.API_AUTHORIZATION_FIELD
    config_url = f"{issuer}/.well-known/openid-configuration"
    jwks_url = f"{issuer}/jwks"

    configuration = {
        "issuer": issuer,
        "jwks_uri": jwks_url,
    }

    keys = {"keys": [rsa_key.public_key_jwk]}

    now = datetime.datetime.now()
    expire = now + datetime.timedelta(days=14)

    jwt_data = {
        "iss": issuer,
        "aud": audience,
        "sub": str(user.uuid),
        "iat": int(now.timestamp()),
        "exp": int(expire.timestamp()),
        auth_field: scopes,
    }
    encoded_jwt = jwt.encode(
        jwt_data, key=rsa_key.private_key_pem, algorithm=rsa_key.jose_algorithm
    )

    requests_mock.get(config_url, json=configuration)
    requests_mock.get(jwks_url, json=keys)

    auth_header = f"{api_token_auth_settings.AUTH_SCHEME} {encoded_jwt}"

    return auth_header


def test_get_profile_information_from_gdpr_api(
    api_client, profile, snapshot, requests_mock, settings
):
    profile.user.first_name = "First"
    profile.user.save()
    auth_header = get_api_token_for_user_with_scopes(
        profile.user, [settings.GDPR_API_QUERY_SCOPE], requests_mock
    )
    api_client.credentials(HTTP_AUTHORIZATION=auth_header)
    response = api_client.get(
        reverse("helsinki_gdpr:gdpr_v1", kwargs={"pk": profile.id})
    )

    assert response.status_code == 200
    snapshot.assert_match(response.json())


@pytest.mark.parametrize("true_value", TRUE_VALUES)
def test_delete_profile_dry_run_data(
    true_value, api_client, profile, requests_mock, settings
):
    auth_header = get_api_token_for_user_with_scopes(
        profile.user, [settings.GDPR_API_DELETE_SCOPE], requests_mock
    )
    api_client.credentials(HTTP_AUTHORIZATION=auth_header)
    response = api_client.delete(
        reverse("helsinki_gdpr:gdpr_v1", kwargs={"pk": profile.id}),
        data={"dry_run": true_value},
        format="json",
    )

    assert response.status_code == 204
    assert Profile.objects.count() == 1
    assert User.objects.count() == 1


@pytest.mark.parametrize("true_value", TRUE_VALUES)
def test_delete_profile_dry_run_query_params(
    true_value, api_client, profile, requests_mock, settings
):
    auth_header = get_api_token_for_user_with_scopes(
        profile.user, [settings.GDPR_API_DELETE_SCOPE], requests_mock
    )
    api_client.credentials(HTTP_AUTHORIZATION=auth_header)
    response = api_client.delete(
        reverse("helsinki_gdpr:gdpr_v1", kwargs={"pk": profile.id})
        + f"?dry_run={true_value}",
    )

    assert response.status_code == 204
    assert Profile.objects.count() == 1
    assert User.objects.count() == 1


def test_delete_profile(api_client, profile, requests_mock, settings):
    auth_header = get_api_token_for_user_with_scopes(
        profile.user, [settings.GDPR_API_DELETE_SCOPE], requests_mock
    )
    api_client.credentials(HTTP_AUTHORIZATION=auth_header)
    response = api_client.delete(
        reverse("helsinki_gdpr:gdpr_v1", kwargs={"pk": profile.id})
    )

    assert response.status_code == 204
    assert Profile.objects.count() == 0
    assert User.objects.count() == 0


@pytest.mark.parametrize("false_value", FALSE_VALUES)
def test_delete_profile_dry_run_query_params_false(
    false_value, api_client, profile, requests_mock, settings
):
    auth_header = get_api_token_for_user_with_scopes(
        profile.user, [settings.GDPR_API_DELETE_SCOPE], requests_mock
    )
    api_client.credentials(HTTP_AUTHORIZATION=auth_header)
    response = api_client.delete(
        reverse("helsinki_gdpr:gdpr_v1", kwargs={"pk": profile.id})
        + f"?dry_run={false_value}",
    )

    assert response.status_code == 204
    assert Profile.objects.count() == 0
    assert User.objects.count() == 0


@pytest.mark.parametrize("false_value", FALSE_VALUES)
def test_delete_profile_dry_run_data_false(
    false_value, api_client, profile, requests_mock, settings
):
    auth_header = get_api_token_for_user_with_scopes(
        profile.user, [settings.GDPR_API_DELETE_SCOPE], requests_mock
    )
    api_client.credentials(HTTP_AUTHORIZATION=auth_header)
    response = api_client.delete(
        reverse("helsinki_gdpr:gdpr_v1", kwargs={"pk": profile.id}),
        data={"dry_run": false_value},
        format="json",
    )

    assert response.status_code == 204
    assert Profile.objects.count() == 0
    assert User.objects.count() == 0


def test_gdpr_api_requires_authentication(api_client, profile):
    response = api_client.get(
        reverse("helsinki_gdpr:gdpr_v1", kwargs={"pk": profile.id})
    )
    assert response.status_code == 401

    response = api_client.delete(
        reverse("helsinki_gdpr:gdpr_v1", kwargs={"pk": profile.id})
    )
    assert response.status_code == 401


def test_user_can_only_access_his_own_profile(
    api_client, user, profile, requests_mock, settings
):
    auth_header = get_api_token_for_user_with_scopes(
        user,
        [settings.GDPR_API_QUERY_SCOPE, settings.GDPR_API_DELETE_SCOPE],
        requests_mock,
    )
    api_client.credentials(HTTP_AUTHORIZATION=auth_header)

    response = api_client.get(
        reverse("helsinki_gdpr:gdpr_v1", kwargs={"pk": profile.id})
    )
    assert response.status_code == 403

    response = api_client.delete(
        reverse("helsinki_gdpr:gdpr_v1", kwargs={"pk": profile.id})
    )
    assert response.status_code == 403


@pytest.mark.parametrize("use_correct_scope", [True, False])
def test_gdpr_query_requires_correct_scope(
    use_correct_scope, api_client, profile, requests_mock, settings
):
    if use_correct_scope:
        auth_header = get_api_token_for_user_with_scopes(
            profile.user,
            [settings.GDPR_API_QUERY_SCOPE, settings.GDPR_API_DELETE_SCOPE],
            requests_mock,
        )
    else:
        auth_header = get_api_token_for_user_with_scopes(
            profile.user, [settings.GDPR_API_DELETE_SCOPE], requests_mock
        )
    api_client.credentials(HTTP_AUTHORIZATION=auth_header)

    response = api_client.get(
        reverse("helsinki_gdpr:gdpr_v1", kwargs={"pk": profile.id})
    )

    if use_correct_scope:
        assert response.status_code == 200
    else:
        assert response.status_code == 403


@pytest.mark.parametrize("use_correct_scope", [True, False])
def test_gdpr_delete_requires_correct_scope(
    use_correct_scope, api_client, profile, requests_mock, settings
):
    if use_correct_scope:
        auth_header = get_api_token_for_user_with_scopes(
            profile.user,
            [settings.GDPR_API_QUERY_SCOPE, settings.GDPR_API_DELETE_SCOPE],
            requests_mock,
        )
    else:
        auth_header = get_api_token_for_user_with_scopes(
            profile.user, [settings.GDPR_API_QUERY_SCOPE], requests_mock
        )
    api_client.credentials(HTTP_AUTHORIZATION=auth_header)

    response = api_client.delete(
        reverse("helsinki_gdpr:gdpr_v1", kwargs={"pk": profile.id})
    )

    if use_correct_scope:
        assert response.status_code == 204
    else:
        assert response.status_code == 403
        assert Profile.objects.count() == 1
        assert User.objects.count() == 1



from django.apps import apps
from django.conf import settings
from django.db import DatabaseError, transaction
from django.shortcuts import get_object_or_404
from helusers.oidc import ApiTokenAuthentication
from rest_framework import serializers, status
from rest_framework.exceptions import APIException
from rest_framework.permissions import IsAuthenticated
from rest_framework.renderers import JSONRenderer
from rest_framework.response import Response
from rest_framework.views import APIView

from helsinki_gdpr.models import SerializableMixin


class DeletionNotAllowed(APIException):
    status_code = 403
    default_detail = "Profile cannot be deleted."
    default_code = "deletion_not_allowed"


class DryRunException(Exception):
    """Indicate that request is being done as a dry run."""


class DryRunSerializer(serializers.Serializer):
    dry_run = serializers.BooleanField(required=False, default=False)


class GDPRScopesPermission(IsAuthenticated):
    def has_permission(self, request, view):
        authenticated = super().has_permission(request, view)
        if authenticated:
            if request.method == "GET":
                return request.auth.has_api_scopes(settings.GDPR_API_QUERY_SCOPE)
            elif request.method == "DELETE":
                return request.auth.has_api_scopes(settings.GDPR_API_DELETE_SCOPE)
        return False

    def has_object_permission(self, request, view, obj):
        if obj.user:
            return request.user == obj.user
        return False


class GDPRAPIView(APIView):
    """Fetch or delete all information related to the profile."""

    renderer_classes = [JSONRenderer]
    authentication_classes = [ApiTokenAuthentication]
    permission_classes = [GDPRScopesPermission]

    def get_object(self) -> SerializableMixin:
        model = apps.get_model(settings.GDPR_API_MODEL)
        obj = get_object_or_404(model, pk=self.kwargs["pk"])
        self.check_object_permissions(self.request, obj)
        return obj

    def check_dry_run(self):
        """Check if parameters provided to the view indicate it's being used for dry_run."""
        data = DryRunSerializer(data=self.request.data)
        query = DryRunSerializer(data=self.request.query_params)
        data.is_valid()
        query.is_valid()

        if data.validated_data["dry_run"] or query.validated_data["dry_run"]:
            raise DryRunException()

    def get(self, request, *args, **kwargs):
        """Retrieve all profile data related to the given id."""
        return Response(self.get_object().serialize(), status=status.HTTP_200_OK)

    def delete(self, request, *args, **kwargs):
        """Delete all data related to the given profile.

        Deletes all data related to the given profile id, or just checks if the data can be deleted,
        depending on the `dry_run` parameter. Raises DeletionNotAllowed if the item

        Dry run delete is expected to always give the same end result as the proper delete i.e. if
        dry run indicated deleting is OK, the proper delete should be OK too.
        """
        try:
            with transaction.atomic():
                obj = self.get_object()
                user = obj.user
                obj.delete()
                user.delete()
                self.check_dry_run()
        except DryRunException:
            # Deletion is possible. Due to dry run, transaction is rolled back.
            pass
        except DatabaseError:
            raise DeletionNotAllowed()

        return Response(status=status.HTTP_204_NO_CONTENT)



SECRET_KEY = "secret"

DATABASES = {"default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"}}

INSTALLED_APPS = (
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "helusers.apps.HelusersConfig",
    "helusers.apps.HelusersAdminConfig",
    "helsinki_gdpr",
    "tests",
)

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "tests.urls"

AUTH_USER_MODEL = "tests.User"

GDPR_API_MODEL = "tests.Profile"
GDPR_API_QUERY_SCOPE = "testprefix.gdprquery"
GDPR_API_DELETE_SCOPE = "testprefix.gdprdelete"

DEBUG = True
USE_TZ = True

OIDC_API_TOKEN_AUTH = {
    "AUDIENCE": "test_audience",
    "ISSUER": "https://test_issuer_1",
    "REQUIRE_API_SCOPE_FOR_AUTHENTICATION": False,
    "API_AUTHORIZATION_FIELD": "",
    "API_SCOPE_PREFIX": "",
}



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots["test_model_serialization 1"] = {
    "children": [
        {"key": "MEMO", "value": "Memo"},
        {"key": "USER", "value": "First"},
        {
            "children": [
                {"children": [{"key": "DATA", "value": "Extra"}], "key": "EXTRADATA"}
            ],
            "key": "EXTRA_DATA",
        },
    ],
    "key": "PROFILE",
}



from tests.factories import ExtraDataFactory


def test_model_serialization(profile, snapshot):
    profile.user.first_name = "First"
    profile.user.save()
    ExtraDataFactory(profile=profile)

    serialized_profile = profile.serialize()

    snapshot.assert_match(serialized_profile)



import uuid

from django.conf import settings
from django.db import models
from helusers.models import AbstractUser

from helsinki_gdpr.models import SerializableMixin


class User(AbstractUser):
    pass


class Profile(SerializableMixin):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    user = models.OneToOneField(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    memo = models.CharField(max_length=128, blank=True)

    serialize_fields = (
        {"name": "memo"},
        {"name": "user", "accessor": lambda x: x.first_name},
        {"name": "extra_data"},
    )


class ExtraData(SerializableMixin):
    profile = models.ForeignKey(
        Profile,
        on_delete=models.CASCADE,
        related_name="extra_data",
    )
    data = models.CharField(max_length=255)

    serialize_fields = ({"name": "data"},)
